@@ -1,6 +1,6 @@
 /* FAudio - XAudio Reimplementation for FNA
  *
- * Copyright (c) 2011-2021 Ethan Lee, Luigi Auriemma, and the MonoGame Team
+ * Copyright (c) 2011-2022 Ethan Lee, Luigi Auriemma, and the MonoGame Team
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -477,13 +477,68 @@ static float APF_OUT_DELAYS[REVERB_COUNT_APF_OUT] =
 	7.73f
 };
 
-static const float STEREO_SPREAD[4] =
+typedef enum FAudio_ChannelPositionFlags
 {
-	0.0f,
-	0.5216f,
-	0.0f,
-	0.5216f
-};
+	Position_Left = 0x1,
+	Position_Right = 0x2,
+	Position_Center = 0x4,
+	Position_Rear = 0x8,
+} FAudio_ChannelPositionFlags;
+
+static FAudio_ChannelPositionFlags FAudio_GetChannelPositionFlags(int32_t total_channels, int32_t channel)
+{
+	switch (total_channels)
+	{
+		case 1:
+			return Position_Center;
+
+		case 2:
+			return (channel == 0) ? Position_Left : Position_Right;
+
+		case 4:
+			switch (channel)
+			{
+				case 0:
+					return Position_Left;
+				case 1:
+					return Position_Right;
+				case 2:
+					return Position_Left | Position_Rear;
+				case 3:
+					return Position_Right | Position_Rear;
+			}
+			FAudio_assert(0 && "Unsupported channel count");
+			break;
+		case 5:
+			switch (channel)
+			{
+				case 0:
+					return Position_Left;
+				case 1:
+					return Position_Right;
+				case 2:
+					return Position_Center;
+				case 3:
+					return Position_Left | Position_Rear;
+				case 4:
+					return Position_Right | Position_Rear;
+			}
+			FAudio_assert(0 && "Unsupported channel count");
+			break;
+		default:
+			FAudio_assert(0 && "Unsupported channel count");
+			break;
+	}
+
+	/* shouldn't happen, but default to left speaker */
+	return Position_Left;
+}
+
+float FAudio_GetStereoSpreadDelayMS(int32_t total_channels, int32_t channel)
+{
+	FAudio_ChannelPositionFlags flags = FAudio_GetChannelPositionFlags(total_channels, channel);
+	return (flags & Position_Right) ? 0.5216f : 0.0f;
+}
 
 typedef struct DspReverbChannel
 {
@@ -503,7 +558,7 @@ typedef struct DspReverb
 	int32_t in_channels;
 	int32_t out_channels;
 	int32_t reverb_channels;
-	DspReverbChannel channel[4];
+	DspReverbChannel channel[5];
 
 	float early_gain;
 	float reverb_gain;
@@ -521,7 +576,7 @@ static inline void DspReverb_Create(
 ) {
 	int32_t i, c;
 
-	FAudio_assert(in_channels == 1 || in_channels == 2);
+	FAudio_assert(in_channels == 1 || in_channels == 2 || in_channels == 6);
 	FAudio_assert(out_channels == 1 || out_channels == 2 || out_channels == 6);
 
 	FAudio_zero(reverb, sizeof(DspReverb));
@@ -538,7 +593,14 @@ static inline void DspReverb_Create(
 		);
 	}
 
-	reverb->reverb_channels = (out_channels == 6) ? 4 : out_channels;
+	if (out_channels == 6)
+	{
+		reverb->reverb_channels = (in_channels == 6) ? 5 : 4;
+	}
+	else
+	{
+		reverb->reverb_channels = out_channels;
+	}
 
 	for (c = 0; c < reverb->reverb_channels; c += 1)
 	{
@@ -554,7 +616,7 @@ static inline void DspReverb_Create(
 			DspCombShelving_Initialize(
 				&reverb->channel[c].lpf_comb[i],
 				sampleRate,
-				COMB_DELAYS[i] + STEREO_SPREAD[c],
+				COMB_DELAYS[i] + FAudio_GetStereoSpreadDelayMS(reverb->reverb_channels, c),
 				500,
 				500,
 				-6,
@@ -569,7 +631,7 @@ static inline void DspReverb_Create(
 			DspAllPass_Initialize(
 				&reverb->channel[c].apf_out[i],
 				sampleRate,
-				APF_OUT_DELAYS[i] + STEREO_SPREAD[c],
+				APF_OUT_DELAYS[i] + FAudio_GetStereoSpreadDelayMS(reverb->reverb_channels, c),
 				0.5f,
 				pMalloc
 			);
@@ -634,13 +696,6 @@ static inline void DspReverb_SetParameters(
 	FAudioFXReverbParameters *params
 ) {
 	float early_diffusion, late_diffusion;
-	float channel_delay[4] =
-	{
-		0.0f,
-		0.0f,
-		params->RearDelay,
-		params->RearDelay
-	};
 	DspCombShelving *comb;
 	int32_t i, c;
 
@@ -662,9 +717,14 @@ static inline void DspReverb_SetParameters(
 	/* Reverberation */
 	for (c = 0; c < reverb->reverb_channels; c += 1)
 	{
+		float channel_delay =
+			(FAudio_GetChannelPositionFlags(reverb->reverb_channels, c) & Position_Rear) ?
+			params->RearDelay :
+			0.0f;
+
 		DspDelay_Change(
 			&reverb->channel[c].reverb_delay,
-			(float) params->ReverbDelay + channel_delay[c]
+			(float) params->ReverbDelay + channel_delay
 		);
 
 		for (i = 0; i < REVERB_COUNT_COMB; i += 1)
@@ -674,7 +734,7 @@ static inline void DspReverb_SetParameters(
 			/* Set decay time of comb filter */
 			DspDelay_Change(
 				&comb->comb_delay,
-				COMB_DELAYS[i] + STEREO_SPREAD[c]
+				COMB_DELAYS[i] + FAudio_GetStereoSpreadDelayMS(reverb->reverb_channels, c)
 			);
 			comb->comb_feedback_gain = DspComb_FeedbackFromRT60(
 				&comb->comb_delay,
@@ -709,11 +769,14 @@ static inline void DspReverb_SetParameters(
 
 	for (c = 0; c < reverb->reverb_channels; c += 1)
 	{
+		FAudio_ChannelPositionFlags position = FAudio_GetChannelPositionFlags(reverb->reverb_channels, c);
+		float gain;
+
 		for (i = 0; i < REVERB_COUNT_APF_OUT; i += 1)
 		{
 			DspAllPass_Change(
 				&reverb->channel[c].apf_out[i],
-				APF_OUT_DELAYS[i] + STEREO_SPREAD[c],
+				APF_OUT_DELAYS[i] + FAudio_GetStereoSpreadDelayMS(reverb->reverb_channels, c),
 				late_diffusion
 			);
 		}
@@ -726,24 +789,39 @@ static inline void DspReverb_SetParameters(
 			params->RoomFilterMain + params->RoomFilterHF
 		);
 
-		reverb->channel[c].gain = 1.5f - (
-			((c % 2 == 0 ?
-				params->PositionMatrixLeft :
-				params->PositionMatrixRight
-			) / 27.0f) * 0.5f
-		);
-		if (c >= 2)
+		if (position & Position_Left)
+		{
+			gain = params->PositionMatrixLeft;
+		}
+		else if (position & Position_Right)
+		{
+			gain = params->PositionMatrixRight;
+		}
+		else /*if (position & Position_Center) */
+		{
+			gain = (params->PositionMatrixLeft + params->PositionMatrixRight) / 2.0f;
+		}
+		reverb->channel[c].gain = 1.5f - (gain / 27.0f) * 0.5f;
+
+		if (position & Position_Rear)
 		{
 			/* Rear-channel Attenuation */
 			reverb->channel[c].gain *= 0.75f;
 		}
 
-		reverb->channel[c].early_gain = 1.2f - (
-			((c % 2 == 0 ?
-				params->PositionLeft :
-				params->PositionRight
-			) / 6.0f) * 0.2f
-		);
+		if (position & Position_Left)
+		{
+			gain = params->PositionLeft;
+		}
+		else if (position & Position_Right)
+		{
+			gain = params->PositionRight;
+		}
+		else /*if (position & Position_Center) */
+		{
+			gain = (params->PositionLeft + params->PositionRight) / 2.0f;
+		}
+		reverb->channel[c].early_gain = 1.2f - (gain / 6.0f) * 0.2f;
 		reverb->channel[c].early_gain = (
 			reverb->channel[c].early_gain *
 			reverb->early_gain
@@ -937,35 +1015,35 @@ static inline float DspReverb_INTERNAL_Process_2_to_2(
 	size_t sample_count
 ) {
 	const float *in_end = samples_in + sample_count;
-	float in, in_ratio, early, late[2];
+	float in, early, late[2];
 	float squared_sum = 0;
 
 	while (samples_in < in_end)
 	{
 		/* Input - Combine 2 channels into 1 */
 		in = (samples_in[0] + samples_in[1]) / 2.0f;
-		in_ratio = in * reverb->dry_ratio;
-		samples_in += 2;
 
 		/* Early Reflections */
 		early = DspReverb_INTERNAL_ProcessEarly(reverb, in);
 
 		/* Reverberation with Wet/Dry Mix */
-		late[0] = DspReverb_INTERNAL_ProcessChannel(
+		late[0] = (DspReverb_INTERNAL_ProcessChannel(
 			reverb,
 			&reverb->channel[0],
 			early
-		);
+		) * reverb->wet_ratio) + samples_in[0] * reverb->dry_ratio;
 		late[1] = (DspReverb_INTERNAL_ProcessChannel(
 			reverb,
 			&reverb->channel[1],
 			early
-		) * reverb->wet_ratio) + in_ratio;
+		) * reverb->wet_ratio) + samples_in[1] * reverb->dry_ratio;
 		squared_sum += (late[0] * late[0]) + (late[1] * late[1]);
 
 		/* Output */
 		*samples_out++ = late[0];
 		*samples_out++ = late[1];
+
+		samples_in += 2;
 	}
 
 	return squared_sum;
@@ -1015,6 +1093,52 @@ static inline float DspReverb_INTERNAL_Process_2_to_5p1(
 	return squared_sum;
 }
 
+static inline float DspReverb_INTERNAL_Process_5p1_to_5p1(
+	DspReverb *reverb,
+	float *restrict samples_in,
+	float *restrict samples_out,
+	size_t sample_count
+) {
+	const float *in_end = samples_in + sample_count;
+	float in, in_ratio, early, late[5];
+	float squared_sum = 0;
+	int32_t c;
+
+	while (samples_in < in_end)
+	{
+		/* Input - Combine non-LFE channels into 1 */
+		in = (samples_in[0] + samples_in[1] + samples_in[2] +
+				samples_in[4] + samples_in[5]) / 5.0f;
+		in_ratio = in * reverb->dry_ratio;
+
+		/* Early Reflections */
+		early = DspReverb_INTERNAL_ProcessEarly(reverb, in);
+
+		/* Reverberation with Wet/Dry Mix */
+		for (c = 0; c < 5; c += 1)
+		{
+			late[c] = (DspReverb_INTERNAL_ProcessChannel(
+				reverb,
+				&reverb->channel[c],
+				early
+			) * reverb->wet_ratio) + in_ratio;
+			squared_sum += late[c] * late[c];
+		}
+
+		/* Output */
+		*samples_out++ = late[0];	/* Front Left */
+		*samples_out++ = late[1];	/* Front Right */
+		*samples_out++ = late[2];	/* Center */
+		*samples_out++ = samples_in[3];	/* LFE, pass through */
+		*samples_out++ = late[3];	/* Rear Left */
+		*samples_out++ = late[4];	/* Rear Right */
+
+		samples_in += 6;
+	}
+
+	return squared_sum;
+}
+
 #undef OUTPUT_SAMPLE
 
 /* Reverb FAPO Implementation */
@@ -1148,7 +1272,8 @@ uint32_t FAudioFXReverb_IsInputFormatSupported(
 	else if (pOutputFormat->nChannels == 6)
 	{
 		if (	pRequestedInputFormat->nChannels != 1 &&
-			pRequestedInputFormat->nChannels != 2	)
+			pRequestedInputFormat->nChannels != 2 &&
+			pRequestedInputFormat->nChannels != 6	)
 		{
 			SET_SUPPORTED_FIELD(nChannels, 1);
 		}
@@ -1200,6 +1325,13 @@ uint32_t FAudioFXReverb_IsOutputFormatSupported(
 			SET_SUPPORTED_FIELD(nChannels, pInputFormat->nChannels);
 		}
 	}
+	else if (pInputFormat->nChannels == 6)
+	{
+		if (pRequestedOutputFormat->nChannels != 6)
+		{
+			SET_SUPPORTED_FIELD(nChannels, pInputFormat->nChannels);
+		}
+	}
 	else
 	{
 		SET_SUPPORTED_FIELD(nChannels, 1);
@@ -1252,7 +1384,9 @@ uint32_t FAudioFXReverb_LockForProcess(
 			 pOutputLockedParameters->pFormat->nChannels == 6)) ||
 		(pInputLockedParameters->pFormat->nChannels == 2 &&
 			(pOutputLockedParameters->pFormat->nChannels == 2 ||
-			 pOutputLockedParameters->pFormat->nChannels == 6))))
+			 pOutputLockedParameters->pFormat->nChannels == 6)) ||
+		(pInputLockedParameters->pFormat->nChannels == 6 &&
+			pOutputLockedParameters->pFormat->nChannels == 6)))
 	{
 		return FAPO_E_FORMAT_UNSUPPORTED;
 	}
@@ -1273,6 +1407,22 @@ uint32_t FAudioFXReverb_LockForProcess(
 		fapo->base.pMalloc
 	);
 
+	/* Initialize the effect to a default setting */
+	if (fapo->apiVersion == 9)
+	{
+		DspReverb_SetParameters9(
+			&fapo->reverb,
+			(FAudioFXReverbParameters9*) fapo->base.m_pParameterBlocks
+		);
+	}
+	else
+	{
+		DspReverb_SetParameters(
+			&fapo->reverb,
+			(FAudioFXReverbParameters*) fapo->base.m_pParameterBlocks
+		);
+	}
+
 	/* Call	parent to do basic validation */
 	return FAPOBase_LockForProcess(
 		&fapo->base,
@@ -1295,7 +1445,7 @@ static inline void FAudioFXReverb_CopyBuffer(
 		return;
 	}
 	
-	/* 1 -> 1 or 2 -> 2 */
+	/* equal channel count */
 	if (fapo->inBlockAlign == fapo->outBlockAlign)
 	{
 		FAudio_memcpy(
@@ -1353,6 +1503,24 @@ void FAudioFXReverb_Process(
 	FAudioFXReverbParameters *params;
 	uint8_t update_params = FAPOBase_ParametersChanged(&fapo->base);
 	float total;
+
+	params = (FAudioFXReverbParameters*) FAPOBase_BeginProcess(&fapo->base);
+
+	/* Update parameters before doing anything else  */
+	if (update_params)
+	{
+		if (fapo->apiVersion == 9)
+		{
+			DspReverb_SetParameters9(
+				&fapo->reverb,
+				(FAudioFXReverbParameters9*) params
+			);
+		}
+		else
+		{
+			DspReverb_SetParameters(&fapo->reverb, params);
+		}
+	}
 	
 	/* Handle disabled filter */
 	if (IsEnabled == 0)
@@ -1369,6 +1537,7 @@ void FAudioFXReverb_Process(
 			);
 		}
 
+		FAPOBase_EndProcess(&fapo->base);
 		return;
 	}
 	
@@ -1382,24 +1551,6 @@ void FAudioFXReverb_Process(
 		);
 	}
 
-	params = (FAudioFXReverbParameters*) FAPOBase_BeginProcess(&fapo->base);
-
-	/* Update parameters  */
-	if (update_params)
-	{
-		if (fapo->apiVersion == 9)
-		{
-			DspReverb_SetParameters9(
-				&fapo->reverb,
-				(FAudioFXReverbParameters9*) params
-			);
-		}
-		else
-		{
-			DspReverb_SetParameters(&fapo->reverb, params);
-		}
-	}
-
 	/* Run reverb effect */
 	#define PROCESS(pin, pout) \
 		DspReverb_INTERNAL_Process_##pin##_to_##pout( \
@@ -1421,10 +1572,14 @@ void FAudioFXReverb_Process(
 			{
 				total = PROCESS(1, 5p1);
 			}
-			else
+			else if (fapo->reverb.in_channels == 2)
 			{
 				total = PROCESS(2, 5p1);
 			}
+			else /* 5.1 */
+			{
+				total = PROCESS(5p1, 5p1);
+			}
 			break;
 	}
 	#undef PROCESS
@@ -1528,16 +1683,6 @@ uint32_t FAudioCreateReverbWithCustomAllocatorEXT(
 		sizeof(FAudioFXReverbParameters) * 3
 	);
 	result->apiVersion = 7;
-	#define INITPARAMS(offset) \
-		FAudio_memcpy( \
-			params + sizeof(FAudioFXReverbParameters) * offset, \
-			&fxdefault, \
-			sizeof(FAudioFXReverbParameters) \
-		);
-	INITPARAMS(0)
-	INITPARAMS(1)
-	INITPARAMS(2)
-	#undef INITPARAMS
 
 	/* Initialize... */
 	FAudio_memcpy(
@@ -1573,6 +1718,13 @@ uint32_t FAudioCreateReverbWithCustomAllocatorEXT(
 	result->base.Destructor = FAudioFXReverb_Free;
 	#undef ASSIGN_VT
 
+	/* Prepare the default parameters */
+	result->base.base.Initialize(
+		result,
+		&fxdefault,
+		sizeof(FAudioFXReverbParameters)
+	);
+
 	/* Finally. */
 	*ppApo = &result->base.base;
 	return 0;
@@ -1701,16 +1853,6 @@ uint32_t FAudioCreateReverb9WithCustomAllocatorEXT(
 		sizeof(FAudioFXReverbParameters9) * 3
 	);
 	result->apiVersion = 9;
-	#define INITPARAMS(offset) \
-		FAudio_memcpy( \
-			params + sizeof(FAudioFXReverbParameters9) * offset, \
-			&fxdefault, \
-			sizeof(FAudioFXReverbParameters9) \
-		);
-	INITPARAMS(0)
-	INITPARAMS(1)
-	INITPARAMS(2)
-	#undef INITPARAMS
 
 	/* Initialize... */
 	FAudio_memcpy(
@@ -1746,6 +1888,13 @@ uint32_t FAudioCreateReverb9WithCustomAllocatorEXT(
 	result->base.Destructor = FAudioFXReverb_Free;
 	#undef ASSIGN_VT
 
+	/* Prepare the default parameters */
+	result->base.base.Initialize(
+		result,
+		&fxdefault,
+		sizeof(FAudioFXReverbParameters9)
+	);
+
 	/* Finally. */
 	*ppApo = &result->base.base;
 	return 0;
